// $Id: file_path.cxx 4812 2007-01-30 00:54:20Z jz $

#include "zschimmer.h"
#include "file.h"
#include "file_path.h"
#include "log.h"
#include <sys/stat.h>

namespace zschimmer {

//-----------------------------------------------------------------------------File_path::File_path
    
File_path::File_path( const string& directory, const string& tail_path ) 
{ 
    if( int len = directory.length() + tail_path.length() )
    {
        if( directory != ""  &&  *directory.rbegin() != Z_DIR_SEPARATOR[0]  &&  *directory.rbegin() != '/' )
        {
            reserve( len + 1 );
            *this = directory;
            *this += directory_separator( *this );
        }
        else
        {
            reserve( len );
            *this = directory;
        }

        if( tail_path != "" )
        {
            if( directory != ""  &&  ( *tail_path.begin() == Z_DIR_SEPARATOR[0]  ||  *tail_path.begin() == '/' ) )
            {
                *this += tail_path.c_str() + 1;
            }
            else
            {
                *this += tail_path;
            }
        }
    }
}

//-------------------------------------------------------------------File_path::directory_separator
    
char File_path::directory_separator( const string& example_path )
{
    if( example_path.find( '/' ) != string::npos )  return '/';
    return Z_DIR_SEPARATOR[0];
}

//-------------------------------------------------------------------File_path::directory_separator
    
char File_path::directory_separator( const string& example_path1, const string& example_path2 )
{
    if( example_path1.find( '/' ) != string::npos )  return '/';
    if( example_path2.find( '/' ) != string::npos )  return '/';
    return Z_DIR_SEPARATOR[0];
}

//----------------------------------------------------------------------File_path::is_absolute_path

bool File_path::is_absolute_path() const
{
    bool result = false;

#   if defined Z_WINDOWS

        if( length() >= 3  
         && isalpha( (unsigned char)(*this)[0] )
         && (*this)[ 1 ] == ':' 
         && ( (*this)[ 2 ] == '/' || (*this)[2] == '\\' )  )  result = true;          // "c:\"

        else

        if( length() >= 2  
         && ( (*this)[0] == '/' || (*this)[0] == '\\' )  
         && ( (*this)[1] == '/' || (*this)[1] == '\\' ) )  result = true;           // \\host\share

#   endif

    if( !result )  result = length() >= 1  &&  ( (*this)[0] == '/' || (*this)[0] == Z_DIR_SEPARATOR[0] );       // "/x" gilt auch unter Windows absolut

    return result;
}

//------------------------------------------------------------------------------File_path::set_name
    
void File_path::set_name( const string& name )
{
    string dir = directory();
    if( dir != "" )  dir += directory_separator( dir );
    set_path( dir + name );
}

//-------------------------------------------------------------------------File_path::set_directory

void File_path::set_directory( const string& directory_path )
{
    set_path( File_path( directory_path, name() ) );
}

//---------------------------------------------------------------------File_path::prepend_directory

void File_path::prepend_directory( const string& directory_path )
{
    set_path( File_path( directory_path, path() ) );
    //string path = directory_path;

    //if( path != ""  &&  ( *path.rbegin() == Z_DIR_SEPARATOR[0]  ||  *path.rbegin() == '/' ) )   // directory_path ohne '/' am Ende?
    //{
    //    path += directory_separator( directory_path, *this );                                   // '/' anfgen
    //}

    //if( *this != ""  &&  ( *rbegin() == Z_DIR_SEPARATOR[0]  ||  *rbegin() == '/' ) )            // Name mit '/' am Anfang?
    //{
    //    erase( 0, 1 );                                                                          // '/' entfernen
    //}

    //set_path( path + *this );
}

//-----------------------------------------------------------------------------File_path::directory

File_path File_path::directory() const
{
    return directory_of_path( path() );
}

//----------------------------------------------------------------------------------File_path::name

string File_path::name() const
{
    return filename_of_path( path() );
}

//-----------------------------------------------------------------------------File_path::base_name

string File_path::base_name() const
{
    return basename_of_path( path() );
}

//-----------------------------------------------------------------------------File_path::extension

string File_path::extension() const
{
    return extension_of_path( path() );
}

//--------------------------------------------------------------------------------File_path::unlink

void File_path::unlink() const
{
    z_unlink( path() );
}

//----------------------------------------------------------------------------File_path::try_unlink

bool File_path::try_unlink() const
{
    int err = ::unlink( c_str() );
    return err == 0;
}

//-------------------------------------------------------------------------------File_path::move_to

void File_path::move_to( const File_path& destination_path ) const
{
    File_path d = destination_path;
    if( d.directory() != ""  &&  d.name() == "" )  d.set_name( name() );

    int err = rename( c_str(), d.c_str() );
    int errn = errno;

#   ifdef Z_WINDOWS
        if( err && errno == EEXIST )
        {
            d.try_unlink();
            err = rename( c_str(), d.c_str() );
            errn = errno;
        }
#   endif

    if( Log_ptr log = "" )
    {
        log << "rename(\"" << *this << "\",\"" << d << ")";
        if( err )  log << "  ERRNO-" << errn << " " << z_strerror( errn );
        log << '\n';
    }

    if( err )  
    {
        throw_errno( errno, "rename", c_str(), d.c_str() );
    }
}

//-----------------------------------------------------------------------------------test_file_path

void test_file_path()
{
    {
        File_path p ( "directory", "name" );
        assert( p == "directory" Z_DIR_SEPARATOR "name" );
    }

    {
        File_path p ( "directory", "/name" );
        assert( p == "directory" Z_DIR_SEPARATOR "name" );
    }

    {
        File_path p ( "directory/", "name" );
        assert( p == "directory/name" );
    }

    {
        File_path p ( "directory/", "/name" );
        assert( p == "directory/name" );
    }

    {
        File_path p ( "directory", "name1/name2" );
        assert( p == "directory" Z_DIR_SEPARATOR "name1/name2" );
    }

    {
        File_path p ( "directory", "/name1/name2" );
        assert( p == "directory" Z_DIR_SEPARATOR "name1/name2" );
    }

    {
        File_path p ( "directory/", "name1/name2" );
        assert( p == "directory/name1/name2" );
    }

    {
        File_path p ( "directory/", "/name1/name2" );
        assert( p == "directory/name1/name2" );
    }

    {
        File_path p;
        p.set_directory( "directory" );
        assert( p == "directory" Z_DIR_SEPARATOR );
    }

    {
        File_path p;
        p.set_directory( "directory/" );
        assert( p == "directory/" );
    }

    {
        File_path p ( "directory", "" );
        assert( p == "directory" Z_DIR_SEPARATOR );
    }

    {
        File_path p ( "directory", "" );
        p.set_name( "name1/name2" );
        assert( p == "directory" Z_DIR_SEPARATOR "name1/name2" );
    }

    {
        File_path p ( "dir1/dir2/name" );
        p.set_directory( "directory" );
        assert( p == "directory" Z_DIR_SEPARATOR "name" );
    }

    {
        File_path p ( "/dir1/dir2/name" );
        p.set_directory( "directory" );
        assert( p == "directory" Z_DIR_SEPARATOR "name" );
    }
}

//-------------------------------------------------------------------------File_info::assert_uint32

void File_info::assert_uint32( time_t t, const char* function )
{
    assert( t <= UINT_MAX );
    if( t > UINT_MAX )  z::throw_xc( function, string_gmt_from_time_t( t ) );
}

//-----------------------------------------------------------------------File_info::set_create_time

void File_info::set_create_time( time_t t )
{ 
    assert_uint32( t, __FUNCTION__ );

    _create_time        = (uint32)t;  
    _create_time_filled = true; 
}

//------------------------------------------------------------------File_info::set_last_access_time

void File_info::set_last_access_time( time_t t )
{ 
    assert_uint32( t, __FUNCTION__ );

    _last_access_time        = (uint32)t;  
    _last_access_time_filled = true; 
}

//-------------------------------------------------------------------File_info::set_last_write_time

void File_info::set_last_write_time( time_t t )
{ 
    assert_uint32( t, __FUNCTION__ );

    _last_write_time        = (uint32)t;  
    _last_write_time_filled = true; 
}

//---------------------------------------------------------------------------File_info::create_time

time_t File_info::create_time()
{
    if( !_create_time_filled )  call_stat(),  assert( _create_time_filled );
    return _create_time;
}

//----------------------------------------------------------------------File_info::last_access_time

time_t File_info::last_access_time()
{
    if( !_last_access_time_filled )  call_stat(),  assert( _last_access_time_filled );
    return _last_access_time;
}

//-----------------------------------------------------------------------File_info::last_write_time

time_t File_info::last_write_time() 
{
    if( !_last_write_time_filled )  call_stat(),  assert( _last_write_time_filled );
    return _last_write_time;
}

//-----------------------------------------------------------------------------File_info::call_stat

void File_info::call_stat()
{
    bool ok = try_call_stat();
    if( !ok )  throw_errno( errno, "stat", _file_path.c_str() );
}

//-------------------------------------------------------------------------File_info::try_call_stat

bool File_info::try_call_stat()
{
    bool        result  = false;
    struct stat stat_buf;

    memset( &stat_buf, 0, sizeof stat_buf );

    int error = stat( _file_path.c_str(), &stat_buf );
    if( !error )
    {
        set_create_time     ( stat_buf.st_ctime );
        set_last_write_time ( stat_buf.st_mtime );
        set_last_access_time( stat_buf.st_atime );
        result = true;
    }
    else
    if( errno != 2 )  throw_errno( errno, "stat", _file_path.c_str() );

    assert( result || errno );
    return result;
}

//--------------------------------------------------------------File_info::quick_compare_last_write

bool File_info::quick_last_write_less( const File_info* a, const File_info* b )
{ 
    time_t at = a->_last_write_time;     // Muss gefllt sein!
    time_t bt = b->_last_write_time;     // Muss gefllt sein!

    if( at < bt )  return true;
    if( at > bt )  return false;

    if( a->_file_path.path() < b->_file_path.path() )  return true;

    return false;
}

//--------------------------------------------------------------File_info::quick_last_write_compare

int File_info::quick_last_write_compare( const File_info* a, const File_info* b )
{ 
    time_t at = a->_last_write_time;     // Muss gefllt sein!
    time_t bt = b->_last_write_time;     // Muss gefllt sein!

    if( at < bt )  return -1;
    if( at > bt )  return +1;

    string a_path = a->_file_path.path();   // name() ist vermutlich langsamer als path()
    string b_path = b->_file_path.path();

    if( a_path < b_path )  return -1;
    if( a_path > b_path )  return +1;

    return 0;
}

//-------------------------------------------------------------------------------------------------

} //namespace zschimmer
