// $Id: z_posix_mutex.cxx 3546 2005-04-03 08:30:29Z jz $

#include "zschimmer.h"
#include "z_posix.h"
#include "mutex.h"
#include "threads.h"


#ifdef Z_UNIX

#include <unistd.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/socket.h>

#ifdef Z_SOLARIS
#   include <stropts.h>
#   include <sys/filio.h>       // ioctl( , FIONBIO )
#   define PTHREAD_MUTEX_FAST_NP  PTHREAD_MUTEX_RECURSIVE    // Solaris kennt PTHREAD_MUTEX_FAST_NP nicht
#endif

#ifdef Z_HPUX
#   define PTHREAD_MUTEX_FAST_NP PTHREAD_MUTEX_ERRORCHECK    // HP kennt PTHREAD_MUTEX_FAST_NP nicht
#endif

//-------------------------------------------------------------------------------------------------

namespace zschimmer {
namespace posix {

//--------------------------------------------------------------------------------------enter_mutex

void enter_mutex( pthread_mutex_t* mutex )
{
    int err = pthread_mutex_lock( mutex );
    if( err )  throw_errno( err, "pthread_mutex_lock" );
}

//--------------------------------------------------------------------------------------leave_mutex

void leave_mutex( pthread_mutex_t* mutex )
{
    int err = pthread_mutex_unlock( mutex );
    if( err )  throw_errno( err, "pthread_mutex_unlock" );
}

//-------------------------------------------------------------------------------------Mutex::Mutex

Mutex::Mutex( const string& name, Kind kind )
:
    Mutex_base(name,kind)
{
    int                 err;
    pthread_mutexattr_t attr;

    memset( &_system_mutex, 0, sizeof _system_mutex );

    err = pthread_mutexattr_init( &attr);
    if( err )  throw_xc( "pthread_mutexattr_init", _name.c_str() );


    int k = (int)PTHREAD_MUTEX_RECURSIVE;
    if( kind == kind_nonrecursive )
    {
//#       ifdef Z_DEBUG
            k = (int)PTHREAD_MUTEX_ERRORCHECK;
//#        else
//            kind = (int)PTHREAD_MUTEX_FAST_NP;
//#       endif
    }

    err = pthread_mutexattr_settype( &attr, k );
    if( err )  throw_xc( "pthread_mutexattr_settype", _name.c_str() );

    err = pthread_mutex_init( &_system_mutex, &attr );
    if( err )  throw_xc( "pthread_mutex_init", _name.c_str() );

    err = pthread_mutexattr_destroy( &attr );
    if( err )  throw_xc( "pthread_mutexattr_destroy", _name.c_str() );
}

//------------------------------------------------------------------------------------Mutex::~Mutex

Mutex::~Mutex()
{
    int err = pthread_mutex_destroy( &_system_mutex );
    if( err )  Z_LOG( "pthread_mutex_init err=" << err << '\n' );
}

//-------------------------------------------------------------------------------------Mutex::enter

void Mutex::enter()
{
    int err = pthread_mutex_lock( &_system_mutex );
    if( err )  throw_errno( err, "pthread_mutex_lock", _name.c_str() );
}

//---------------------------------------------------------------------------------Mutex::try_enter

bool Mutex::try_enter()
{
    int err = pthread_mutex_trylock( &_system_mutex );
    if( err ) 
    {
        if( err == EBUSY )  return false;
        throw_errno( err, "pthread_mutex_lock", _name.c_str() );
    }

    return true;
}

//-------------------------------------------------------------------------------------Mutex::leave

void Mutex::leave()
{
    int err = pthread_mutex_unlock( &_system_mutex );
    if( err )  throw_errno( err, "pthread_mutex_unlock", _name.c_str() );
}

//-------------------------------------------------------------------------Mutex::locking_thread_id

Thread_id Mutex::locking_thread_id()
{
    return 0;   // Sollte die Thread-Id des sperrenden Threads sein.
}

//-------------------------------------------------------------------------------------------------

} //namespace posix
} //namespace zschimmer

#endif
